package tutorial

import com.nativelibs4java.opencl._

case class Interval(offset: Int, stride: Int, length: Int) {
  
  def /(parts: Int): Array[Interval] = {
    val newLength = length / parts
    val newStride = stride * parts
    (0 until parts).map(i => Interval(offset + stride * i, newStride, newLength))(collection.breakOut)
  }
}
trait Op
case class Op1(i: Interval) extends Op
case class Op2(o1: Op, o2: Op) extends Op
case class Op4(o1: Op, o2: Op, o3: Op, o4: Op) extends Op

object FFT {
  
  def computeOp(i: Interval): Op = {
    if ((i.length & 3) == 0) { // i.length is a multiple of 4
      val Array(o1, o2, o3, o4) = (i / 4) map computeOp
      Op4(o1, o2, o3, o4)
    } else if ((i.length & 1) == 0) { // i.length is odd
      val Array(o1, o2) = (i / 2) map computeOp
      Op2(o1, o2)
    } else {
      Op1(i)
    }
  }
  
  
  def main(args: Array[String]): Unit = {
    implicit val context = JavaCL.createBestContext(CLPlatform.DeviceFeature.DoubleSupport)
    implicit val queue = context.createDefaultOutOfOrderQueueIfPossible
    val fft = new FFT
  }
}

class FFT(implicit context: CLContext, queue: CLQueue) {
  val program = new FFTProgram(context)
  
  import FFT._
  
  def execute(op: Op, in: CLDoubleBuffer, outputLength: Int, forward: Boolean, eventsToWaitFor: CLEvent*): (CLDoubleBuffer, CLEvent) = {
    val out = context.createDoubleBuffer(CLMem.Usage.InputOutput, outputLength)
    val sign = if (forward) 1 else -1
    val evt = op match {
      case Op1(i) =>
        program.dft(queue, in, out, i.offset, i.stride, i.length, outputLength, sign, Array(outputLength), null, eventsToWaitFor:_*)
      case Op2(o1, o2) =>
        val (out1, evt1) = execute(o1, in, outputLength, forward, eventsToWaitFor:_*)
        val (out2, evt2) = execute(o2, in, outputLength, forward, eventsToWaitFor:_*)
        program.fft_radix2(queue, out1, out2, out, outputLength, sign, Array(outputLength), null, evt1, evt2)
      case Op4(o1, o2, o3, o4) =>
        val (out1, evt1) = execute(o1, in, outputLength, forward, eventsToWaitFor:_*)
        val (out2, evt2) = execute(o2, in, outputLength, forward, eventsToWaitFor:_*)
        val (out3, evt3) = execute(o3, in, outputLength, forward, eventsToWaitFor:_*)
        val (out4, evt4) = execute(o4, in, outputLength, forward, eventsToWaitFor:_*)
        program.fft_radix4(queue, out1, out2, out3, out4, out, outputLength, sign, Array(outputLength), null, evt1, evt2, evt3, evt4)
    }
    (out, evt)
  }
  def fft(in: Array[Double], forward: Boolean): Array[Double] = {
    import java.nio._
    val (outBuf, outEvt) = fft(context.createDoubleBuffer(CLMem.Usage.Input, DoubleBuffer.wrap(in), true), forward)
    val nioBuf = outBuf.read(queue, outEvt)
    assert(nioBuf.capacity == in.length)
    val out = new Array[Double](in.length)
    nioBuf.get(out)
    out
  }
  def fft(in: CLDoubleBuffer, forward: Boolean, eventsToWaitFor: CLEvent*): (CLDoubleBuffer, CLEvent) = {
    val length = in.getElementCount.toInt
    val op = computeOp(Interval(0, 1, length))
    println("OP = " + op)
    
    execute(op, in, length, forward, eventsToWaitFor:_*)
  }
}
/*

x x x x x x x x Interval(0, 1, 8)

x   x   x   x   Interval(0, 2, 4)
  x   x   x   x Interval(1, 2, 4)
  
x       x       Interval(0, 4, 2)
    x       x   Interval(2, 4, 2)
  x       x     Interval(1, 4, 2)
      x       x Interval(3, 4, 2)

*/
